# MIP Model

# Objetivo: Minimizar los costos operativos de la logística de última milla
# y la selección de flota óptima considerando ELV.
# Se toma en cuenta la optimización de ruteo.

# Autores: Sebastian Zazueta, Roberto Galarza y Vicente Montemayor
# Institución: Universidad de Monterrey
# email: sebastian.zazueta@udem.edu, roberto.galarza@udem.edu,
#        vicente.montemayor@udem.edu

#Importar librerias 
    
import collections
from ortools.linear_solver import pywraplp
import time
import numpy as np
import pandas as pd
   
#Factor para escalar parametro
factor = 1000

#Parámetros

#COMPLETAR INFORMACIÓN (Características Paquetes Peso y Volumen)
p_w = (np.array([1.2,.04])*factor).tolist() # Peso en kilos; Volumen en m3
p_w = list(map(int,p_w))
 
base_dir = r'C:\Users\vicen\Desktop\UDEM\PEF\Matrices'

#Arcos disponibles para circular de LEV
r_kijLEV = pd.read_excel(base_dir+'\R_kij_centra_LEV.xlsx')
r_kijLEV = (np.array(r_kijLEV)).tolist()

#Arcos disponibles para circular de Motos y Trucks
r_kijMT = pd.read_excel(base_dir+'\R_kij_centra_MT.xlsx')
r_kijMT = (np.array(r_kijMT)).tolist()

#Arcos con ciclovías
b_ij = pd.read_excel(base_dir+'\Bij_Centralizado.xlsx')
b_ij = (np.array(b_ij)).tolist()

#Distancias de arcos ij
l_ij = pd.read_excel(base_dir+'\Lij_Centralizado_1000.xlsx')
l_ij.replace({1000:1},inplace=True)
l_ij = (np.array(l_ij)*factor).tolist()

#Demanda en nodos i
d_i = pd.read_excel(base_dir+'\di_cs_baja.xlsx')
d_i = (np.array(d_i)).tolist()[0]


#Costo fijo de vehículo por día
cf_k = [22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544, 22.544,
        28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3, 28.3,
        17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165, 17.165,
        47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481, 47.481]
cf_k = (np.array(cf_k)*factor).tolist()

#Costo variable de vehículo por día
cv_k = [0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001,
        0.004,  0.004, 0.004, 0.004, 0.004, 0.004,  0.004, 0.004, 0.004, 0.004, 0.004,  0.004, 0.004,
        0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029, 0.029,
        0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223, 0.223]
cv_k = (np.array(cv_k)*factor).tolist()

#Capacidad maxima por vehiculo
Cmax_kw = [
           [100,0.196],
           [100,0.196], 
           [100,0.196], 
           [100,0.196], 
           [100,0.196], #e-bike
           [275,0.957],
           [275,0.957],
           [275,0.957],
           [275,0.957],
           [275,0.957],#e-trike
           [40,0.15], 
           [40,0.15],
           [40,0.15],#moto
           [900,10],
           [900,10]  #truck
          ]

#Escalamiento de cmax
Cmax_kw = (np.array(Cmax_kw)*factor).tolist()
Cmax_kw = [list(map(int,l)) for l in Cmax_kw]

#Capacidad de distancia por vehiculo k en km
Dmax_k = [30,30,30,30,30,30,30,30,30,30,30,30,30,
          30,30,30,30,30,30,30,30,30,30,30,30,30,
          500,500,500,500,500,500,500,500,500,500,500,500, 500,
          500,500,500,500,500,500,500,500,500,500,500,500,500]
Dmax_k = (np.array(Dmax_k)*factor).tolist()

#tiempo de ruta máximo por vehículo
Tmax_k = [8,8,8,8,8,8,8,8,8,8,8,8,8,
          8,8,8,8,8,8,8,8,8,8,8,8,8,
          8,8,8,8,8,8,8,8,8,8,8,8,8,
          8,8,8,8,8,8,8,8,8,8,8,8,8,]
Tmax_k = (np.array(Tmax_k)*factor).tolist()

#Capacidad de energía por vehículo
Emax_k = [56,56,56,56,56,56,56,56,56,56,56,56,56,
          98,98,98,98,98,98,98,98,98,98,98,98,98,
          20,20,20,20,20,20,20,20,20,20,20,20,20,
          95,95,95,95,95,95,95,95,95,95,95,95,95]
Emax_k = (np.array(Emax_k)*factor).tolist()

#Energy consumption rate per vehicle
e_k = [0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,0.55,
       3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,3.36,
       0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,
       0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,]
e_k = (np.array(e_k)).tolist()

#Velocidad promedio por vehiculo k
s_k = [20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20]
s_k = (np.array(s_k)*factor).tolist()

#Velocidad adicional por ir en ciclovia
sa_k = [5,5,5,5,5,0,0,0,0,0,0,0,0,0,0]
sa_k = (np.array(sa_k)*factor).tolist()

#Tiempo de servicio Fijo por vehículo
stf_k = [0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,
         0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,0.01,
         0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,0.012,
         0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28,0.28]
stf_k = (np.array(stf_k)*factor).tolist()

#Tiempo de servicio variable por vehículo
stv_k = [0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,
         0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,
         0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,0.083,
         0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133,0.133]
stv_k = (np.array(stv_k)*factor).tolist()

#Indirect variables
n_vehicles = len(cf_k)
n_nodes = len(l_ij)
n_features = len(p_w)

#Sets
Vt = [0] #Transhipment depot
VaLEV=[52,59,60,61,64,65]
VaMT = [52,59,60,61,64,65]
Vc = []
Vn = [i for i in range(n_nodes) if i not in Vt]

for i,d in enumerate(d_i):
    if d>0:
        Vc.append(i)
        
#---------------------------------------------------------------
#SOLVER   
solver = pywraplp.Solver('VRP',
                         pywraplp.Solver.SAT_INTEGER_PROGRAMMING)
#BIG M
M=100000

#---------------------------------------------------------------
#VARIABLES

X = collections.defaultdict(lambda: collections.defaultdict(lambda: collections.defaultdict(dict)))
Xd = collections.defaultdict(lambda: collections.defaultdict(lambda: collections.defaultdict(dict)))
O = collections.defaultdict(lambda: collections.defaultdict(dict))
H = collections.defaultdict(lambda: collections.defaultdict(dict))
U = collections.defaultdict(lambda: collections.defaultdict(lambda: collections.defaultdict(dict)))

# ijk Variables
for i in range(n_nodes):
    for j in range(n_nodes):     
        for k in range(n_vehicles):  
            
            if k<=9 and r_kijLEV[i][j]==1:
                X[i][j][k] = solver.BoolVar('')  #Includes constriaint 15 
                Xd[i][j][k] = 1
            elif k>9 and r_kijMT[i][j]==1:
                X[i][j][k] = solver.BoolVar('')
                Xd[i][j][k] = 1
            else:
                X[i][j][k] =  0#solver.BoolVar('')
                Xd[i][j][k] = 0

#Vehicle-node variables
for k in range(n_vehicles): 
    
    for i in range(n_nodes): 
        
        if i in Vc:
            #Node i is visited by vehicle k
            O[i][k] = solver.BoolVar('') #Includes constraint 15
        
        #Sequence of vehicle k for visiting node i
        H[i][k] = solver.IntVar(0,n_nodes,'') #Includes constraint 14
        
        for w in range(n_features):
            
            #Remaining cargo vehicle k in node i regarding feature w
            U[i][k][w] = solver.IntVar(0,Cmax_kw[k][w],'') #Includes constraint 13, 14
         
#----------------------------------------------------------------
#CONSTRAINTS

for k in range(n_vehicles):
    
    solver.Add( sum([X[0][j][k] for j in range(n_nodes)]) <= 1 ) #1
    solver.Add( sum([X[i][0][k] for i in range(n_nodes)]) <= 1 ) #2
       
    if k <= 9:
        solver.Add( M * sum([X[0][j][k] for j in VaLEV]) >= sum([O[i][k] for i in Vc]) ) #3
        solver.Add( M * sum([X[i][0][k] for i in VaLEV]) >= sum([O[i][k] for i in Vc]) ) #4
    if k>9:
        solver.Add( M * sum([X[0][j][k] for j in VaMT]) >= sum([O[i][k] for i in Vc]) ) #3
        solver.Add( M * sum([X[i][0][k] for i in VaMT]) >= sum([O[i][k] for i in Vc]) ) #4

    for j in range(n_nodes):     
        solver.Add( sum([X[i][j][k] for i in range(n_nodes)]) == sum([X[j][i][k] for i in range(n_nodes)]) ) #6

    for i in Vc: 
        solver.Add( sum([X[i][j][k] for j in range(n_nodes)]) >= O[i][k]) #7
       
    #Escalar
    solver.Add( sum([X[i][j][k]*l_ij[i][j]*e_k[k] for i in range(n_nodes) for j in range(n_nodes)]) <= Emax_k[k] ) #9
    solver.Add( sum([X[i][j][k]*l_ij[i][j] for i in range(n_nodes) for j in range(n_nodes)]) <= Dmax_k[k] ) #10
    solver.Add( sum([ (l_ij[i][j] / (s_k[k]+sa_k[k]*b_ij[i][j]) ) * X[i][j][k] for i in range(n_nodes) for j in range(n_nodes)])
                + sum([ (stf_k[k] + stv_k[k]*d_i[i])*O[i][k] for i in Vc]) 
               <= Tmax_k[k] ) #11
    

for i in Vc:
    
    solver.Add(sum([ O[i][k] for k in range(n_vehicles) ]) == 1) #5
    
    for j in range(n_nodes):
         for k in range(n_vehicles):
             for w in range(n_features):
                
                 #Escalar
                 solver.Add( U[j][k][w] <= U[i][k][w] - d_i[i]*p_w[w]*O[i][k] + M*(1-X[i][j][k])) #8

for i in Vn:
    for j in Vn:
        for k in range(n_vehicles):
            
            solver.Add( H[i][k] - H[j][k] + (n_nodes)*X[i][j][k] <= n_nodes-1) #12

       
#----------------------------------------------------------------
#OBJECTIVE FUNCTION
cv = sum([X[i][j][k]*cv_k[k]*l_ij[i][j] for i in range(n_nodes) for j in range(n_nodes) for k in range(n_vehicles)])
cf = sum([X[0][j][k]*cf_k[k] for j in VaLEV for k in range(10)]) + sum([X[0][j][k]*cf_k[k] for j in VaMT for k in range(10,n_vehicles)])
fo = cv + cf

#----------------------------------------------------------------
#SOLVE
    
solver.Minimize(fo)
solver.SetTimeLimit(1800*1000)

start=time.time()
status = solver.Solve()
end=time.time()


#----------------------------------------------------------------
#RESULT
import pandas as pd

#Guardamos matrices en lista de dfs para visualizar
final=[]
final_o= []
for veh in range(n_vehicles):
    resultado=[]
    resultado_o = []
    for origen in range(n_nodes):
        result=[]
        if origen in O:
            resultado_o.append(O[origen][veh].solution_value())
        else:
            resultado_o.append(0)
        for destino in range(n_nodes):
            
            #result.append(X[origen][destino][bus].solution_value()) 
            
            if veh <= 9:  
            
                if r_kijLEV[origen][destino]==1:
                    result.append(X[origen][destino][veh].solution_value()) 
                else:
                    result.append(0)
        
            if veh > 9:
            
                if r_kijMT[origen][destino]==1:
                    result.append(X[origen][destino][veh].solution_value())   
                else:
                    result.append(0)
                         
        resultado.append(result)
    resultado_final=pd.DataFrame(resultado) 
    final.append(resultado_final)
    final_o.append(resultado_o)

routes=collections.defaultdict(list)
for k in range(n_vehicles):
    origin=0
    routes[k].append(0)
    for t in range(int(sum([X[i][j][k].solution_value() for j in range(n_nodes) for i in range(n_nodes) if i!=j and Xd[i][j][k] == 1]))):
        for j in range(n_nodes):
            
            # if X[origin][j][k].solution_value()==1:
            #         routes[k].append(j)
            #         origin=j
            #         break
            
            if k <= 9:  
            
                if r_kijLEV[origin][j] and X[origin][j][k].solution_value()==1:
                    routes[k].append(j)
                    origin=j
                    break
        
            if k > 9:
            
                if r_kijMT[origin][j] and X[origin][j][k].solution_value()==1:
                    routes[k].append(j)
                    origin=j
                    break
            

if status == pywraplp.Solver.OPTIMAL:
    print("Optimal solution found.")
elif status==pywraplp.Solver.FEASIBLE:
    print("Feasible solution found.")
else: 
    print("Infeasible")
obj = solver.Objective().Value()
print("Total CPU time: "+ str(round(end-start,4)) +" seconds.")
print('Objetive =', obj)
print("--------------------------------------------------------")

